//===---- LoongArchISelDAGToDAG.h - A Dag to Dag Inst Selector for LoongArch --------===//
//
// Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
// See https://llvm.org/LICENSE.txt for license information.
// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
//
//===----------------------------------------------------------------------===//
//
// This file defines an instruction selector for the LoongArch target.
//
//===----------------------------------------------------------------------===//

#ifndef LLVM_LIB_TARGET_LOONGARCH_LOONGARCHISELDAGTODAG_H
#define LLVM_LIB_TARGET_LOONGARCH_LOONGARCHISELDAGTODAG_H

#include "LoongArch.h"
#include "LoongArchSubtarget.h"
#include "LoongArchTargetMachine.h"
#include "llvm/CodeGen/SelectionDAGISel.h"

//===----------------------------------------------------------------------===//
// Instruction Selector Implementation
//===----------------------------------------------------------------------===//

//===----------------------------------------------------------------------===//
// LoongArchDAGToDAGISel - LoongArch specific code to select LoongArch machine
// instructions for SelectionDAG operations.
//===----------------------------------------------------------------------===//
namespace llvm {

class LoongArchDAGToDAGISel : public SelectionDAGISel {
public:
  explicit LoongArchDAGToDAGISel(LoongArchTargetMachine &TM, CodeGenOpt::Level OL)
                : SelectionDAGISel(TM, OL), Subtarget(nullptr) {}

  // Pass Name
  StringRef getPassName() const override {
    return "LoongArch DAG->DAG Pattern Instruction Selection";
  }

  bool runOnMachineFunction(MachineFunction &MF) override;

  void PostprocessISelDAG() override;

  void getAnalysisUsage(AnalysisUsage &AU) const override;

private:
  /// Keep a pointer to the LoongArchSubtarget around so that we can make the right
  /// decision when generating code for different targets.
  const LoongArchSubtarget *Subtarget;
  // Include the pieces autogenerated from the target description.
  #include "LoongArchGenDAGISel.inc"

  void doPeepholeLoadStoreADDI();

  bool selectAddrFrameIndex(SDValue Addr, SDValue &Base, SDValue &Offset) const;

  bool selectAddrFrameIndexOffset(SDValue Addr, SDValue &Base, SDValue &Offset,
                                  unsigned OffsetBits,
                                  unsigned ShiftAmount) const;

  // Complex Pattern.
  /// (reg + imm).
  bool selectAddrRegImm(SDValue Addr, SDValue &Base, SDValue &Offset) const;

  /// Fall back on this function if all else fails.
  bool selectAddrDefault(SDValue Addr, SDValue &Base, SDValue &Offset) const;

  /// Match integer address pattern.
  bool selectIntAddr(SDValue Addr, SDValue &Base, SDValue &Offset) const;

  bool selectAddrRegImm12(SDValue Addr, SDValue &Base,
                         SDValue &Offset) const;

  /// Match addr+simm12 and addr
  bool selectIntAddrSImm12(SDValue Addr, SDValue &Base,
                           SDValue &Offset) const;

  bool selectIntAddrSImm10(SDValue Addr, SDValue &Base,
                           SDValue &Offset) const;

  bool selectIntAddrSImm10Lsl1(SDValue Addr, SDValue &Base,
                               SDValue &Offset) const;

  bool selectIntAddrSImm10Lsl2(SDValue Addr, SDValue &Base,
                               SDValue &Offset) const;

  bool selectIntAddrSImm9Lsl3(SDValue Addr, SDValue &Base,
                              SDValue &Offset) const;

  bool selectIntAddrSImm11Lsl1(SDValue Addr, SDValue &Base,
                               SDValue &Offset) const;

  bool selectIntAddrSImm14Lsl2(SDValue Addr, SDValue &Base,
                               SDValue &Offset) const;

  bool selectIntAddrSImm10Lsl3(SDValue Addr, SDValue &Base,
                               SDValue &Offset) const;

  /// Select constant vector splats.
  bool selectVSplat(SDNode *N, APInt &Imm, unsigned MinSizeInBits) const;
  /// Select constant vector splats whose value fits in a given integer.
  bool selectVSplatCommon(SDValue N, SDValue &Imm, bool Signed,
                          unsigned ImmBitSize) const;
  /// Select constant vector splats whose value fits in a uimm1.
  bool selectVSplatUimm1(SDValue N, SDValue &Imm) const;
  /// Select constant vector splats whose value fits in a uimm2.
  bool selectVSplatUimm2(SDValue N, SDValue &Imm) const;
  /// Select constant vector splats whose value fits in a uimm3.
  bool selectVSplatUimm3(SDValue N, SDValue &Imm) const;
  /// Select constant vector splats whose value fits in a uimm4.
  bool selectVSplatUimm4(SDValue N, SDValue &Imm) const;
  /// Select constant vector splats whose value fits in a uimm5.
  bool selectVSplatUimm5(SDValue N, SDValue &Imm) const;
  /// Select constant vector splats whose value fits in a uimm6.
  bool selectVSplatUimm6(SDValue N, SDValue &Imm) const;
  /// Select constant vector splats whose value fits in a uimm8.
  bool selectVSplatUimm8(SDValue N, SDValue &Imm) const;
  /// Select constant vector splats whose value fits in a simm5.
  bool selectVSplatSimm5(SDValue N, SDValue &Imm) const;
  /// Select constant vector splats whose value is a power of 2.
  bool selectVSplatUimmPow2(SDValue N, SDValue &Imm) const;
  /// Select constant vector splats whose value is the inverse of a
  /// power of 2.
  bool selectVSplatUimmInvPow2(SDValue N, SDValue &Imm) const;
  /// Select constant vector splats whose value is a run of set bits
  /// ending at the most significant bit
  bool selectVSplatMaskL(SDValue N, SDValue &Imm) const;
  /// Select constant vector splats whose value is a run of set bits
  /// starting at bit zero.
  bool selectVSplatMaskR(SDValue N, SDValue &Imm) const;

  void Select(SDNode *N) override;

  bool trySelect(SDNode *Node);

  // getImm - Return a target constant with the specified value.
  inline SDValue getImm(const SDNode *Node, uint64_t Imm) {
    return CurDAG->getTargetConstant(Imm, SDLoc(Node), Node->getValueType(0));
  }

  bool SelectInlineAsmMemoryOperand(const SDValue &Op,
                                    unsigned ConstraintID,
                                    std::vector<SDValue> &OutOps) override;
};

FunctionPass *createLoongArchISelDag(LoongArchTargetMachine &TM,
                                       CodeGenOpt::Level OptLevel);
}

#endif
